/****************************************************************************
** $Id: xtermcontrol.c,v 1.33 2002/10/26 17:20:36 jet Exp $
**
** Copyright (C) 2002 Jess Thrysoee <jess@thrysoee.dk>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <termios.h>
#include <config.h>

#include "getopt.h"
#include "xtermcontrol.h"
#include "configuration.h"

/* The name the program was run with, stripped of any leading path. */
char *program_name;

static char temp[BUFSIZ];

/* controlling terminal streams and descriptor */
FILE *ctty, *tty_in, *tty_out;
int TTY_FILENO;

struct termios tty_ts, tty_ts_orig; /* termios settings */
struct termios *tty_ts_orig_pt = NULL;

struct controlseqs ctlseqstab[] =
{
   {10,  0, NULL, OSC,                          /* FG           */
         "foreground",
         "--fg=COLOR",
         "set xterm foreground color"},

   {11,  0, NULL, OSC,                          /* BG           */
         "background",
         "--bg=COLOR",
         "set xterm background color"},

   {12,  0, NULL, OSC,                          /* CURSOR       */
         "cursor",
         "--cursor=COLOR",
         "set cursor color"},

   {13,  0, NULL, OSC,                          /* MOUSE_FG     */
         "mouse-foreground",
         "--mouse-fg=COLOR",
         "set mouse pointer foreground color"},

   {14,  0, NULL, OSC,                          /* MOUSE_BG     */
         "mouse-background",
         "--mouse-bg=COLOR",
         "set mouse pointer background color"},

   {50,  0, NULL, OSC,                          /* FONT         */
         "font",
         "--font=FONT",
         "set xterm font"},

   { 0,  0, NULL, OSC,                          /* TITLE        */
         "title",
         "--title=STRING",
         "set xterm's window title"},

   { 8,  3, NULL, CSI,                          /* GEOMETRY     */
         "geometry",
         "--geometry=WIDTHxHEIGHT+XOFF+YOFF",
         "set xterm size and/or position"},

   {10,  0, NULL, GET_OSC,                      /* GET_FG       */
         "foreground",
         "--get-fg",
         "report xterm foreground color"},

   {11,  0, NULL, GET_OSC,                      /* GET_BG       */
         "background",
         "--get-bg",
         "report xterm background color"},

   {12,  0, NULL, GET_OSC,                      /* GET_CURSOR   */
         "cursor",
         "--get-cursor",
         "report cursor color"},

   {13,  0, NULL, GET_OSC,                      /* GET_MOUSE_FG */
         "mouse-foreground",
         "--get-mouse-fg",
         "report mouse pointer foreground color"},

   {14,  0, NULL, GET_OSC,                      /* GET_MOUSE_BG */
         "mouse-background",
         "--get-mouse-bg",
         "report mouse pointer background color"},

   {50,  0, NULL, GET_OSC,                      /* GET_FONT     */
         "font",
         "--get-font",
         "report xterm font"},

   {21,  0, NULL, GET_CSI,                      /* GET_TITLE    */
         "title",
         "--get-title",
         "report xterm's window title"},

   {18, 13, NULL, GET_CSI,                      /* GET_GEOMETRY */
         "geometry",
         "--get-geometry",
         "report xterm size and position"},

   { 9,  1, NULL, CSI, NULL,                    /* MAXIMIZE     */
         "--maximize",
         "maximize window"},

   { 9,  0, NULL, CSI, NULL,                    /* RESTORE      */
         "--restore",
         "restore maximized window"},

   { 2,  0, NULL, CSI, NULL,                    /* ICONIFY      */
         "--iconify",
         "iconify window"},

   { 1,  0, NULL, CSI, NULL,                    /* DE_ICONIFY   */
         "--de-iconify",
         "de-iconify window"},

   { 5,  0, NULL, CSI, NULL,                    /* RAISE        */
         "--raise",
         "raise window"},

   { 6,  0, NULL, CSI, NULL,                    /* LOWER        */
         "--lower",
         "lower window"},

   { 0,  0, NULL, ANY,                          /* RAW          */
         "raw",
         "--raw=CTLSEQS",
         "issue raw control sequence"}
};

int
main(int argc, char** argv)
{
   int i, c, ctl1, ctl2, type;
   int verbose = 0;
   int configuration_missing = 0;
   char* text = NULL;
   char configuration_file[256];

   configuration list;              /* configuration linked list */
   const configuration_element *lp; /* configuration linked list element */

   program_name = argv[0];
   if (program_name && strrchr(program_name, '/'))
        program_name = strrchr(program_name, '/') + 1;

   configuration_file[0] = '\0';

   while(1) {
      static int a = -1;
      int longindex = 0;

      static struct option longopts[] =
      {
         {"fg"           , 1, &a, FG},
         {"bg"           , 1, &a, BG},
         {"cursor"       , 1, &a, CURSOR},
         {"mouse-fg"     , 1, &a, MOUSE_FG},
         {"mouse-bg"     , 1, &a, MOUSE_BG},
         {"font"         , 1, &a, FONT},
         {"title"        , 1, &a, TITLE},
         {"geometry"     , 1, &a, GEOMETRY},
         {"get-fg"       , 0, &a, GET_FG},
         {"get-bg"       , 0, &a, GET_BG},
         {"get-cursor"   , 0, &a, GET_CURSOR},
         {"get-mouse-fg" , 0, &a, GET_MOUSE_FG},
         {"get-mouse-bg" , 0, &a, GET_MOUSE_BG},
         {"get-font"     , 0, &a, GET_FONT},
         {"get-title"    , 0, &a, GET_TITLE},
         {"get-geometry" , 0, &a, GET_GEOMETRY},
         {"maximize"     , 0, &a, MAXIMIZE},
         {"restore"      , 0, &a, RESTORE},
         {"iconify"      , 0, &a, ICONIFY},
         {"de-iconify"   , 0, &a, DE_ICONIFY},
         {"raise"        , 0, &a, RAISE},
         {"lower"        , 0, &a, LOWER},
         {"raw"          , 1, &a, RAW},
         {"file"         , 1,  0, 'f'},
         {"verbose"      , 0,  0, 'v'},
         {"version"      , 0,  0, 'u'},
         {"help"         , 0,  0, 'h'},
         { 0             , 0,  0,  0 }
      };

      c = getopt_long(argc, argv, "+vh",
                      longopts, &longindex);
      if (c == -1)
         break;

      switch(c) {
         case 'u':
            version();

         case 'h':
            usage(EXIT_SUCCESS);

         case 'v':
            verbose = 1;
            break;

         case 'f':
            if (*optarg == '~')
               snprintf(configuration_file, sizeof(configuration_file), "%s%s",
                     getenv("HOME"), optarg+1);
            else
               snprintf(configuration_file, sizeof(configuration_file), "%s", optarg);
            break;

         case 0:
            switch (a) {
               case FG       :
               case BG       :
               case CURSOR   :
               case MOUSE_FG :
               case MOUSE_BG :
               case FONT     :
               case TITLE    :
               case RAW      :
                  ctlseqstab[a].text = optarg;
                  break;
               case GET_FG       :
               case GET_BG       :
               case GET_CURSOR   :
               case GET_MOUSE_FG :
               case GET_MOUSE_BG :
               case GET_FONT     :
               case GET_TITLE    :
               case GET_GEOMETRY :
               case MAXIMIZE     :
               case RESTORE      :
               case ICONIFY      :
               case DE_ICONIFY   :
               case RAISE        :
               case LOWER        :
                  ctlseqstab[a].text = "";
                  break;
               case GEOMETRY:
                  validate_geometry(optarg);
                  ctlseqstab[GEOMETRY].text = optarg;
                  break;
               default:
                  usage(EXIT_FAILURE);
            }
            break;

         case '?':
         case ':':
         default :
            usage(EXIT_FAILURE);
      }
   }

   /* read configuration files */
   configuration_init(&list);
   if (!configuration_file[0]) {
      /* default ~/.program_name configuration file */
      snprintf(configuration_file, sizeof(configuration_file), "%s/.%s",
            getenv("HOME"), program_name);
      if (configuration_read(&list, configuration_file) == -1)
         configuration_missing = 1;
   } else {
      /* user specified configuration file */
      if (configuration_read(&list, configuration_file) == -1) {
         fprintf(stderr, "%s: %s: No such file or directory\n",
               program_name, configuration_file);
         do_exit(EXIT_FAILURE);
      }
   }


   tty_control();

   for (i = 0; i < NSEQ; i++) {
      type = ctlseqstab[i].type;
      text = ctlseqstab[i].text;

      temp[0] = '\0';

      /* check if configuration exists */
      if (!text && (type == OSC || type == CSI)) {
         lp = configuration_find(&list, ctlseqstab[i].conf_title);
         if (lp)
            text = lp->value;
      }

      if (text) {
         ctl1 = ctlseqstab[i].ctl1;
         ctl2 = ctlseqstab[i].ctl2;

         switch(type) {
            case CSI:
               switch(i) {
                  case RESTORE :
                  case MAXIMIZE:
                     csi_print2(ctl1, ctl2);
                     break;
                  case GEOMETRY:
                     set_geometry(ctl1, ctl2, text);
                     break;
                  default:
                     csi_print1(ctl1);
                     break;
               }
               break;

            case GET_CSI:
               switch(i) {
                  case GET_GEOMETRY:
                     /* check if display is available */
                     open_display();
                     get_geometry(temp, sizeof(temp), verbose, ctl1, ctl2);
                     break;
                  case GET_TITLE:
                     get_title(temp, sizeof(temp), verbose, ctl1);
                     break;
                  default:
                     csi_print1(ctl1);
                     tty_read(temp, sizeof(temp), ';');
                     break;
               }
               if (*temp)
                  fprintf(stdout, "%s\n", temp);
               break;

            case OSC:
               osc_print(ctl1, text);
               break;

            case GET_OSC:
               get_osc(temp, sizeof(temp), verbose, i, ctl1);
               if (*temp)
                  fprintf(stdout, "%s\n", temp);
               break;

            case ANY:
               raw_print(text);
               tty_read(temp, sizeof(temp), '\n');
               if (*temp)
                  fprintf(stdout, "%s\n", temp);
               break;

            default:
               fprintf(stderr, "unknown type: %d\n", type);
               exit(EXIT_FAILURE);
         }
      }
   }

   configuration_free(&list);

   if (configuration_missing)
      configuration_write(configuration_file);

   do_exit(EXIT_SUCCESS);

   /* NOTREACHED */

   return 0; /* shut up the compiler */
}

void
tty_control(void)
{
   ctty = NULL;

   if (!isatty(STDIN_FILENO) || !isatty(STDOUT_FILENO)) {
      /* get controlling terminal */
      if ((ctty = fopen(ctty_path(), "r+")) == NULL) {
         fprintf(stderr, "%s: failed to get controlling terminal\n", program_name);
         exit(EXIT_FAILURE);
      }
   }
   if (!isatty(STDIN_FILENO) && ctty) {
      tty_in = ctty;
      TTY_FILENO = fileno(tty_in);
   } else {
      tty_in = stdin;
      TTY_FILENO = STDIN_FILENO;
   }
   if (!isatty(STDOUT_FILENO) && ctty)
      tty_out = ctty;
   else
      tty_out = stdout;

   set_tty_raw();
}

/* find path to controlling terminal */
char*
ctty_path(void)
{
   int i;
   char* tty_path = NULL;

   for (i = 0; i <= 2; i++)
      if ((tty_path = ttyname(i)))
         break;

   return tty_path;
}

void
set_tty_raw(void)
{
   /* get and backup tty_in termios */
   tcgetattr(TTY_FILENO, &tty_ts);
   tty_ts_orig = tty_ts;
   tty_ts_orig_pt = &tty_ts_orig;

   /* set tty raw */
   tty_ts.c_iflag = 0;
   tty_ts.c_lflag = 0;
   tty_ts.c_cc[VMIN] = 0;
   tty_ts.c_cc[VTIME] = 5;
   tty_ts.c_iflag &= ~(ICRNL | IXON);
   tty_ts.c_lflag &= ~(ICANON | ECHO | ECHONL);
   tcsetattr(TTY_FILENO, TCSAFLUSH, &tty_ts);
}

void
set_tty_restore(void)
{
   /* restore tty mode */
   if (tty_ts_orig_pt)
      tcsetattr(TTY_FILENO, TCSAFLUSH, tty_ts_orig_pt);
}

/* issue raw escape sequence */
void
raw_print(char* ctlseq)
{
   int c;

   while ((c = *ctlseq++)) {
      if (c == '\\' && *ctlseq) {
         switch (c = *ctlseq++) {
            case 'e':
               c = '\033';
               break;
            case 'a':
               c = '\007';
               break;
            case 'b':
               c = '\b';
               break;
            case 'f':
               c = '\f';
               break;
            case 'n':
               c = '\n';
               break;
            case 'r':
               c = '\r';
               break;
            case 't':
               c = '\t';
               break;
            case 'v':
               c = (int) 0x0B;
               break;
            case '0':
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
               c -= '0';
               if (*ctlseq >= '0' && *ctlseq <= '7')
                  c = (c * 8) + (*ctlseq++ - '0');
               if (*ctlseq >= '0' && *ctlseq <= '7')
                  c = (c * 8) + (*ctlseq++ - '0');
               break;
            case '\\':
               break;
            default:
               fputc('\\', tty_out);
               break;
         }
      }
      fputc(c, tty_out);
   }

   fflush(tty_out);
}

/* read raw terminal output */
void
tty_read(char* output, size_t size, char esc_subst)
{
   int i = 0;

   /* remove ESC from tty_in */
   read(TTY_FILENO, output, 1);

   while (read(TTY_FILENO, output+i, 1) == 1) {
      if (*(output+i) == '\033' || *(output+i) == '\007')
         *(output+i) = esc_subst;
      if (i++ == size-1) break;
   }
   output[i] = '\0';
}

/* issue CSI (Control Sequence Introducer) escape sequence */
void
csi_print1(int ctl1)
{
   snprintf(temp, sizeof(temp), "\033[%dt", ctl1);
   raw_print(temp);
}

void
csi_print2(int ctl1, int ctl2)
{
   snprintf(temp, sizeof(temp), "\033[%d;%dt", ctl1, ctl2);
   raw_print(temp);
}

void
csi_print3(int ctl1, int ctl2, int ctl3)
{
   snprintf(temp, sizeof(temp), "\033[%d;%d;%dt", ctl1, ctl2, ctl3);
   raw_print(temp);
}

/* issue OSC (Operating System Controls) escape sequence */
void
osc_print(int ps, char* pt)
{
   if (pt && *pt)
      snprintf(temp, sizeof(temp), "\033]%d;%s\007", ps, pt);
   else
      snprintf(temp, sizeof(temp), "\033]%d;?\007", ps);

   raw_print(temp);
}

void
get_title(char* title, size_t size, int verbose, int ctl1)
{
   static char local_temp[BUFSIZ];
   static char local_temp2[BUFSIZ];

   csi_print1(ctl1);
   /* use vertical tab ('\v') as "end of title" mark */
   tty_read(local_temp, sizeof(local_temp), '\v');
   if (sscanf(local_temp, "%*[^l]l%[^\v]s", local_temp2) != 1) {
      fprintf(stderr, "%s: failed to read xterm title\n", program_name);
      do_exit(EXIT_FAILURE);
   }
   if (!verbose)
      snprintf(title, size, "%s", local_temp2);
   else
      snprintf(title, size, "%s=\"%s\"",
            ctlseqstab[GET_TITLE].conf_title, local_temp2);
}

void
get_osc(char* osc, size_t size, int verbose, int option, int ctl1)
{
   static char local_temp[BUFSIZ];
   static char local_temp2[BUFSIZ];

   char* conf_title = ctlseqstab[option].conf_title;

   osc_print(ctl1, NULL);
   tty_read(local_temp, sizeof (local_temp), '\n');
   if (sscanf(local_temp, "%*[^;];%s", local_temp2) != 1) {
      fprintf(stderr, "%s: failed to read %s\n", program_name, conf_title);
      do_exit(EXIT_FAILURE);
   }
   if (!verbose)
      snprintf(osc, size, "%s", local_temp2);
   else
      snprintf(osc, size, "%s=\"%s\"", conf_title, local_temp2);
}

void
usage(int status)
{
   int i;

   if (status) {
      fprintf(stderr, "Usage: %s [OPTIONS]...\n", program_name);
      fprintf(stderr, "Try `%s --help' for more information.\n", program_name);

   } else {
      printf("Usage: %s [OPTIONS]...\n", program_name);
      printf("\n");
      printf("Purpose:\n");
      printf("  This program enables dynamic control of xterm properties.\n");
      printf("\n");
      printf("Options:\n");

      for (i = 0; i < NSEQ; i++) {
         printf("  %-35s %s\n", ctlseqstab[i].synopsis, ctlseqstab[i].description);
      }
      printf("  %-35s %s\n", "--file=FILE", "alternative configuration file");
      printf("  %-35s %s\n", "--verbose, -v", "print verbose reports");
      printf("  %-35s %s\n", "--help, -h", "print this help and exit");
      printf("  %-35s %s\n", "--version", "print the version number and exit");
      printf("\n");
      printf("Report bugs to <jess@thrysoee.dk>\n");
    }

   do_exit(status);
}

void
version(void)
{
   printf("%s %s\n", program_name, VERSION);
   printf("\n");
   printf("Copyright (C) 2002 Jess Thrysoee.\n");
   printf("This is free software; see the source for copying conditions. There is NO\n");
   printf("warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");

   do_exit(EXIT_SUCCESS);
}

int
configuration_write(const char* filepath)
{
   FILE* stream;

   stream = fopen(filepath, "w");
   if (stream == NULL) {
#ifndef NDEBUG
      fprintf(stderr, "could not open stream: %s\n", filepath);
#endif
      return -1;
   }

   fprintf(stream, "#\n");
   fprintf(stream, "# This is the configuration file for xtermcontrol.\n");
   fprintf(stream, "#\n");
   fprintf(stream, "# Each line in the file is either a comment or contains an attribute.\n");
   fprintf(stream, "# Attributes consist of a keyword and an associated value:\n");
   fprintf(stream, "#\n");
   fprintf(stream, "# keyword = value # comment\n");
   fprintf(stream, "#\n");
   fprintf(stream, "# Whitespace is ignored in attributes unless within a quoted value. The\n");
   fprintf(stream, "# character '#' is taken to begin a comment. Each '#' and all remaining\n");
   fprintf(stream, "# characters on that line is ignored.\n");
   fprintf(stream, "#\n");
   fprintf(stream, "\n");
   fprintf(stream, "# xterm foreground color\n");
   fprintf(stream, "#foreground=\"black\"\n");
   fprintf(stream, "\n");
   fprintf(stream, "# xterm background color\n");
   fprintf(stream, "#background=\"white\"\n");
   fprintf(stream, "\n");
   fprintf(stream, "# cursor color\n");
   fprintf(stream, "#cursor=\"black\"\n");
   fprintf(stream, "\n");
   fprintf(stream, "# mouse pointer foreground color\n");
   fprintf(stream, "#mouse-foreground=\"black\"\n");
   fprintf(stream, "\n");
   fprintf(stream, "# mouse pointer background color\n");
   fprintf(stream, "#mouse-background=\"white\"\n");
   fprintf(stream, "\n");
   fprintf(stream, "# xterm size and/or position\n");
   fprintf(stream, "#geometry=\"80x24\"\n");
   fprintf(stream, "\n");
   fprintf(stream, "# xterm font\n");
   fprintf(stream, "#font=\"fixed\"\n");

   fclose(stream);

   return 0;
}

void
do_exit(int status)
{
   set_tty_restore();
   if (ctty)
      fclose(ctty);
   exit(status);
}
