/****************************************************************************
** $Id: xtermcontrol.h,v 1.13 2002/08/01 21:54:17 jet Exp $
**
** Copyright (C) 2002 Jess Thrysoee <jess@thrysoee.dk>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
****************************************************************************/

#ifndef XTERMCONTROL_H
#define XTERMCONTROL_H

#include <X11/Xlib.h>

#define NSEQ (sizeof ctlseqstab / sizeof ctlseqstab[0])

enum CTLSEQS_TYPE
{
   CSI,
   GET_CSI,
   OSC,
   GET_OSC,
   ANY
};

enum OPTIONS
{
   FG,
   BG,
   CURSOR,
   MOUSE_FG,
   MOUSE_BG,
   FONT,
   TITLE,
   GEOMETRY,
   GET_FG,
   GET_BG,
   GET_CURSOR,
   GET_MOUSE_FG,
   GET_MOUSE_BG,
   GET_FONT,
   GET_TITLE,
   GET_GEOMETRY,
   MAXIMIZE,
   RESTORE,
   ICONIFY,
   DE_ICONIFY,
   RAISE,
   LOWER,
   RAW
};

/* xterm control sequences are described in ctlseq.ms, which is 
   distributed with the xterm source package, obtainable from 
   http://dickey.his.com/xterm.
   To read the troff file as plain text, issue
   "groff -ms -t -Tlatin1 ctlseqs.ms | col -b > ctlseqs.txt" */

struct controlseqs {
   int ctl1;           /* Ps */
   int ctl2;           /* Ps */
   char* text;         /* Pt */
   int type;           /* CTLSEQS_TYPE */
   char* conf_title;   /* used in reports and configuration file */
   char* synopsis;     /* used in usage */
   char* description;  /* used in usage and configuration file */
};

/* geometry.c */
extern Display* open_display(void);
extern void validate_geometry(char* geometry);
extern void set_geometry(int ctl1, int ctl2, char* geometry);
extern void get_geometry(char* geometry, size_t size, int verbose, int ctl1, int ctl2);

/* xtermcontrol.c */
extern void csi_print1(int ctl1);
extern void csi_print2(int ctl1, int ctl2);
extern void csi_print3(int ctl1, int ctl2, int ctl3);

extern void osc_print(int ps, char* pt);

extern void tty_control(void);
extern char* ctty_path(void);
extern void raw_print(char* ctlseq);
extern void tty_read(char* output, size_t size, char esc_subst);

extern void set_tty_raw(void);
extern void set_tty_restore(void);

extern void get_title(char* title, size_t size, int verbose, int ctl1);
extern void get_osc(char* osc, size_t size, int verbose, int option, int ctl1);

extern int configuration_write(const char* filepath);

extern void usage(int status);
extern void version(void);

extern void do_exit(int status);


#endif /* XTERMCONTROL_H */
